package chain

import (
	"pump/constant"
	"pump/contract/bonding"
	dbModel "pump/model/db"
	"pump/util"
	"strings"

	"github.com/ethereum/go-ethereum/core/types"
	"github.com/shopspring/decimal"
	log "github.com/sirupsen/logrus"
)

func (s *Sync) FilterLaunch(txLog types.Log) {
	if len(txLog.Topics) == 0 {
		return
	}

	abi, _ := bonding.BondingMetaData.GetAbi()
	if txLog.Topics[0].Hex() != abi.Events["Launched"].ID.Hex() {
		return
	}

	launchEvent, err := s.bondingCa.ParseLaunched(txLog)
	if err != nil {
		log.WithError(err).Error("parse launch log")
		return
	}

	newToken := launchEvent.Token.Hex()

	tokenInfo, err := s.d.GetTokenInfo(newToken)
	if err != nil {
		log.WithField("token", newToken).WithError(err).Error("get token info")
		return
	}

	token := &dbModel.Token{
		Creator:     strings.ToLower(tokenInfo.Creator.Hex()),
		Token:       strings.ToLower(tokenInfo.Token.Hex()),
		NFTAddress:  strings.ToLower(tokenInfo.NftAddress.Hex()),
		Pair:        strings.ToLower(tokenInfo.Pair.Hex()),
		Ticker:      tokenInfo.Data.Ticker,
		Name:        tokenInfo.Data.Name,
		Supply:      decimal.NewFromBigInt(tokenInfo.Data.Supply, 0),
		Description: tokenInfo.Description,
		Image:       tokenInfo.Image,
		Twitter:     tokenInfo.Twitter,
		Website:     tokenInfo.Website,
		Telegram:    tokenInfo.Telegram,
		Graduated:   false,
		Price:       decimal.NewFromBigInt(tokenInfo.Data.Price, 0),
		MarketCap:   decimal.NewFromBigInt(tokenInfo.Data.MarketCap, 0),
		Volume:      decimal.NewFromBigInt(tokenInfo.Data.Volume, 0),
		Volume24H:   decimal.NewFromBigInt(tokenInfo.Data.Volume24H, 0),
		Reserve0:    decimal.NewFromBigInt(constant.TokenSupply, 0),
		Reserve1:    decimal.NewFromBigInt(constant.AssetSupply, 0),
		TxHash:      txLog.TxHash.Hex(),
		BlockNumber: int(txLog.BlockNumber),
		BlockTime:   util.BlockToTimestamp(int(txLog.BlockNumber)),
	}

	err = s.d.CreateToken(token)
	if err != nil {
		log.WithError(err).Error("create agent token")
		return
	}

	log.WithFields(log.Fields{
		"creator": tokenInfo.Creator.Hex(),
		"token":   tokenInfo.Token.Hex(),
		"tx":      txLog.TxHash.Hex(),
	}).Info("launch success")
}

func (s *Sync) FilterGraduate(txLog types.Log) {
	if len(txLog.Topics) == 0 {
		return
	}

	abi, _ := bonding.BondingMetaData.GetAbi()
	if txLog.Topics[0].Hex() != abi.Events["Graduated"].ID.Hex() {
		return
	}

	graduateEvent, err := s.bondingCa.ParseGraduated(txLog)
	if err != nil {
		log.WithError(err).Error("parse graduate log")
		return
	}

	err = s.d.GraduateToken(graduateEvent.Token.Hex(), txLog.TxHash.Hex(), graduateEvent.Univ2Pair.Hex())
	if err != nil {
		log.WithError(err).Error("graduate agent token")
		return
	}

	log.WithFields(log.Fields{
		"token":     graduateEvent.Token.Hex(),
		"tx":        txLog.TxHash.Hex(),
		"univ2pair": graduateEvent.Univ2Pair.Hex(),
	}).Info("graduate success")

}

func (s *Sync) FilterSwap(txLog types.Log) {
	if len(txLog.Topics) == 0 {
		return
	}

	abi, _ := bonding.BondingMetaData.GetAbi()
	if txLog.Topics[0].Hex() != abi.Events["Swap"].ID.Hex() {
		return
	}

	swapEvent, err := s.bondingCa.ParseSwap(txLog)
	if err != nil {
		log.WithError(err).Error("parse swap log")
		return
	}
	buy := true
	amountIn, amountOut := swapEvent.Amount1In, swapEvent.Amount0Out // eth增加， token减少
	assetAmount, TokenAmount := decimal.NewFromBigInt(amountIn, 0), decimal.NewFromBigInt(amountOut, 0)
	if swapEvent.Amount1In.String() == "0" {
		buy = false
		amountIn, amountOut = swapEvent.Amount0In, swapEvent.Amount1Out // eth减少， token增加
		assetAmount, TokenAmount = decimal.NewFromBigInt(amountOut, 0), decimal.NewFromBigInt(amountIn, 0)
	}
	price := assetAmount.DivRound(TokenAmount, 18)
	// priceWei := assetAmount.Div(TokenAmount).Mul(decimal.NewFromInt(1000000000000000000))
	// eth单位
	amountInETH := decimal.NewFromBigInt(amountIn, 0).DivRound(decimal.NewFromInt(1000000000000000000), 18)
	amountOutETH := decimal.NewFromBigInt(amountOut, 0).DivRound(decimal.NewFromInt(1000000000000000000), 18)

	tradeHistory := &dbModel.TradeHistory{
		FromAddress: strings.ToLower(swapEvent.Sender.Hex()),
		Token:       strings.ToLower(swapEvent.Token.Hex()),
		AmountIn:    amountInETH,
		AmountOut:   amountOutETH,
		Price:       price,
		IsBuy:       buy,
		TxHash:      txLog.TxHash.Hex(),
		LogIndex:    int(txLog.Index),
		BlockTime:   util.BlockToTimestamp(int(txLog.BlockNumber)),
		BlockNumber: int(txLog.BlockNumber),
	}

	err = s.d.CreateTradeHistory(tradeHistory)
	if err != nil {
		log.WithError(err).Error("create trade history")
		return
	}

	supplyDec := decimal.NewFromBigInt(constant.TokenSupply, 0)
	reserve0Dec := decimal.NewFromBigInt(swapEvent.Reserve0, 0)
	reserve1Dec := decimal.NewFromBigInt(swapEvent.Reserve1, 0)
	volumeDec := decimal.NewFromBigInt(swapEvent.Volume, 0)
	volume24hDec := decimal.NewFromBigInt(swapEvent.Volume24H, 0)

	mc := supplyDec.Mul(reserve1Dec).DivRound(reserve0Dec, 0)
	err = s.d.UpdateToken(swapEvent.Token.String(), tradeHistory.Price, mc, volumeDec, volume24hDec, reserve0Dec, reserve1Dec)
	if err != nil {
		log.WithError(err).Error("update agent token")
		return
	}

	log.WithFields(log.Fields{
		"sender":    swapEvent.Sender.Hex(),
		"token":     swapEvent.Token.Hex(),
		"amountIn":  amountIn,
		"amountOut": amountOut,
		"isBuy":     buy,
		"tx":        txLog.TxHash.Hex(),
	}).Info("swap success")
}
