package chain

import (
	"flag"
	"os"
	"pump/config"
	"pump/contract/bonding"
	"pump/contract/pair"
	"pump/contract/token"
	"pump/dao"
	"strconv"
	"time"

	"github.com/ethereum/go-ethereum/common"
	log "github.com/sirupsen/logrus"
)

type Sync struct {
	c         *config.Config
	d         *dao.Dao
	heightKey string
	bondingCa *bonding.Bonding
	pairCa    *pair.Pair
	tokenCa   *token.Token
}

var manualSync = flag.Int("sync", 0, "sync block height")

func NewSync(_c *config.Config, _d *dao.Dao) (sync *Sync) {
	if *manualSync == 0 && os.Getenv("SYNC") != "" {
		*manualSync, _ = strconv.Atoi(os.Getenv("SYNC"))
	}

	bondingCa, err := bonding.NewBonding(common.HexToAddress(_c.Chain.BondingContract), nil)
	if err != nil {
		panic(err)
	}

	pairCa, err := pair.NewPair(common.HexToAddress(_c.Chain.BondingContract), nil)
	if err != nil {
		panic(err)
	}

	tokenCa, err := token.NewToken(common.HexToAddress(_c.Chain.BondingContract), nil)
	if err != nil {
		panic(err)
	}

	sync = &Sync{
		c:         _c,
		d:         _d,
		heightKey: "height",
		bondingCa: bondingCa,
		pairCa:    pairCa,
		tokenCa:   tokenCa,
	}

	return sync
}

func (s *Sync) Start() {
	lastHeight, err := s.d.GetStorageHeight(s.heightKey)
	if err != nil {
		log.WithError(err).Error("get last block height")
		return
	}

	if lastHeight != 1 {
		// 数据库里保存的是已完成的区块, 再次同步时+1
		lastHeight++
	}

	if *manualSync > 0 {
		lastHeight = *manualSync
	}

	log.WithField("height", lastHeight).Info("last sync block height")
	var latestHeight int
	var beginHeight = lastHeight
	var endHeight = beginHeight + s.c.Chain.BatchBlock

	for {
		latestHeight, err = s.d.GetBlockHeight(s.c.Chain.BehindBlock)
		if err != nil {
			log.WithError(err).Error("get latest block height")
			time.Sleep(time.Second)
			continue
		}
		if (latestHeight-s.c.Chain.BatchBlock)-beginHeight < s.c.Chain.BatchBlock {
			time.Sleep(4 * time.Second)
			continue
		}
		s.SyncLogs(beginHeight-s.c.Chain.BatchBlock, endHeight-s.c.Chain.BatchBlock)

		if err = s.d.SetStorageHeight(s.heightKey, endHeight); err != nil {
			log.WithError(err).Error("set last block height")
		}

		log.WithFields(log.Fields{
			"begin height":  beginHeight,
			"end height":    endHeight,
			"latest height": latestHeight,
			"diff height":   latestHeight - endHeight,
		}).Info("sync block")
		beginHeight = endHeight + 1
		endHeight = beginHeight + s.c.Chain.BatchBlock
	}
}

func (s *Sync) SyncLogs(beginHeight, endHeight int) {
	if endHeight < 0 {
		return
	}

	if beginHeight < 0 {
		beginHeight = 0
	}

	abi, _ := bonding.BondingMetaData.GetAbi()
	topics := []string{
		abi.Events["Launched"].ID.Hex(),
		abi.Events["Graduated"].ID.Hex(),
		abi.Events["Swap"].ID.Hex(),
	}
	logs, err := s.d.GetLogs(beginHeight, endHeight, topics, []string{s.c.Chain.BondingContract})
	if err != nil {
		log.WithFields(log.Fields{"begin": beginHeight, "end": endHeight}).WithError(err).Error("rpc: get logs")
		return
	}

	for _, txLog := range logs {
		// s.Filterpump(txLog)
		s.FilterLaunch(txLog)
		s.FilterGraduate(txLog)
		s.FilterSwap(txLog)
	}
}
